from .gru import GRU
from .gru_component import GRUBlock
from score_models.layers.conv1dsame import Conv1dSame, ConvTransposed1dSame
from score_models.layers.conv2dsame import Conv2dSame, ConvTransposed2dSame
from score_models.layers.conv3dsame import Conv3dSame, ConvTransposed3dSame 
from score_models.layers.up_or_downsampling import upsample, downsample



import torch
from torch import nn
from torch.nn import functional as F
from .gru import GRU

class GRUBlock(nn.Module):
    """
    Abstraction for the recurrent block inside the RIM with 2 hidden states, mostly following
    the architecture described in Morningstar et al (2019), https://arxiv.org/pdf/1901.01359.pdf
    """
    def __init__(self, in_ch:int, hidden_units:int=32, kernel_size:int=5, dimensions:int=0, conv_kernel_size=11):
        super().__init__()
        self.dimensions = dimensions
        if dimensions == 0:
            layer = lambda in_ch, out_ch: nn.Linear(in_ch, out_ch, bias=False)
        elif dimensions == 1:
            layer = lambda in_ch, out_ch: nn.Conv1d(in_ch, out_ch, conv_kernel_size, padding="same", bias=False)
        elif dimensions == 2:
            layer = lambda in_ch, out_ch: nn.Conv2d(in_ch, out_ch, conv_kernel_size, padding="same", bias=False)
        elif dimensions == 3:
            layer = lambda in_ch, out_ch: nn.Conv3d(in_ch, out_ch, conv_kernel_size, padding="same", bias=False)
        else:
            raise ValueError(f"The dimensions {dimensions} is not supported. Only 0, 1, 2 and 3 dimensional input (not including channels) are.")
        
        if hidden_units % 2 != 0:
            raise ValueError(f"hidden units {hidden_units} must be divisible by 2")
        
        self.gru1 = GRU(in_ch, hidden_units//2, kernel_size, dimensions)
        self.gru2 = GRU(hidden_units//2, hidden_units//2, kernel_size, dimensions)
        self.hidden_layer = layer(hidden_units//2, hidden_units//2)
        self.reshape_layer = layer(hidden_units//2, in_ch)

    def forward(self, x, h):
        h_1, h_2 = torch.tensor_split(h, 2, dim=1)
        h1, _ = self.gru1(x, h_1)
        pre_h2 = F.tanh(self.hidden_layer(h1))
        x, h_2 = self.gru2(pre_h2, h_2)
        x = self.reshape_layer(x)
        h = torch.concat([h_1, h_2], dim=1)
        return x, h




import torch
from torch import nn
from torch.nn import functional as F
from rim.definitions import DTYPE

class GRU(nn.Module):
    """
    Gated Recurrent Unit (GRU) with convolutional layers
    """
    def __init__(self, in_ch:int, hidden_units:int=32, kernel_size:int=5, dimensions:int=0):
        super().__init__()
        self.dimensions = dimensions
        self.bias_broadcast = [1, -1] + [1]*dimensions
        if dimensions == 0:
            layer = lambda in_ch, out_ch: nn.Linear(in_ch, out_ch, bias=False)
        elif dimensions == 1:
            layer = lambda in_ch, out_ch: nn.Conv1d(in_ch, out_ch, kernel_size, padding="same", bias=False)
        elif dimensions == 2:
            layer = lambda in_ch, out_ch: nn.Conv2d(in_ch, out_ch, kernel_size, padding="same", bias=False)
        elif dimensions == 3:
            layer = lambda in_ch, out_ch: nn.Conv3d(in_ch, out_ch, kernel_size, padding="same", bias=False)
        else:
            raise ValueError(f"The dimensions {dimensions} is not supported. Only 0, 1, 2 and 3 dimensional input (not including channels) are.")
        # Update gate
        self.w_z = layer(in_ch, hidden_units)
        self.u_z = layer(hidden_units, hidden_units)
        self.bias_z = nn.Parameter(torch.zeros(hidden_units, dtype=DTYPE))

        # Reset gate
        self.w_r = layer(in_ch, hidden_units)
        self.u_r = layer(hidden_units, hidden_units)
        self.bias_r = nn.Parameter(torch.zeros(hidden_units, dtype=DTYPE))
        
        # Candidate activation gate
        self.w_h = layer(in_ch, hidden_units)
        self.u_h = layer(hidden_units, hidden_units)
        self.bias_h = nn.Parameter(torch.zeros(hidden_units, dtype=DTYPE))

    def forward(self, x, h):
        """
        Compute the new state tensor h_{t} from h_{t-1} and x_{t}
        """
        z = F.sigmoid(self.w_z(x) + self.u_z(h) + self.bias_z.view(1, -1, *[1]*self.dimensions))  # update gate
        r = F.sigmoid(self.w_r(x) + self.u_r(h) + self.bias_r.view(1, -1, *[1]*self.dimensions))  # reset gate
        h_tilde = F.tanh(self.w_h(x) + self.u_h(r * h) + self.bias_h.view(1, -1, *[1]*self.dimensions))  # candidate activation
        new_state = (1 - z) * h + z * h_tilde
        return new_state, new_state  # h_{t+1}



from rim.architectures import Hourglass
import torch
import pytest


def test_hourglass_1d():
    H = 32
    B = 10
    C = 15 # channels
    ch_mult = (2, 2)
    dim = 1
    D = [16]*dim
    x = torch.randn(B, C, *D)
    score = torch.randn(B, C, *D)
    model = Hourglass(C, ch_mult=ch_mult, dimensions=dim)
    h = model.initalize_hidden_states(D, B)
    assert h.shape == torch.Size([B, H, *[d//4 for d in D]])
    print(h.shape)
    g, new_h = model(x, None, score, h)
    assert g.shape == torch.Size([B, C, *D])
    assert new_h.shape == h.shape

@pytest.mark.skip(reason="This will not work, I would need to know in advance the correct padding")
def test_hourglass_odd_input_size():
    H = 32
    B = 10
    C = 15 # channels
    ch_mult = (2, 2)
    dim = 1
    D = [15]*dim
    x = torch.randn(B, C, *D)
    score = torch.randn(B, C, *D)
    model = Hourglass(C, ch_mult=ch_mult, dimensions=dim)
    h = model.initalize_hidden_states(D, B)
    assert h.shape == torch.Size([B, H, *[d//4 for d in D]])
    print(h.shape)
    g, new_h = model(x, None, score, h)
    assert g.shape == torch.Size([B, C, *D])
    assert new_h.shape == h.shape

        
def test_hourglass_2d():
    H = 32
    B = 10
    C = 15 # channels
    ch_mult = (2, 2)
    dim = 2 
    D = [16]*dim
    x = torch.randn(B, C, *D)
    score = torch.randn(B, C, *D)
    model = Hourglass(C, ch_mult=ch_mult, dimensions=dim)
    h = model.initalize_hidden_states(D, B)
    assert h.shape == torch.Size([B, H, *[d//4 for d in D]])
    print(h.shape)
    g, new_h = model(x, None, score, h)
    assert g.shape == torch.Size([B, C, *D])
    assert new_h.shape == h.shape
        
def test_hourglass_3d():
    H = 32
    B = 10
    C = 15 # channels
    ch_mult = (2, 2)
    dim =3
    D = [16]*dim
    x = torch.randn(B, C, *D)
    score = torch.randn(B, C, *D)
    model = Hourglass(C, ch_mult=ch_mult, dimensions=dim)
    h = model.initalize_hidden_states(D, B)
    assert h.shape == torch.Size([B, H, *[d//4 for d in D]])
    print(h.shape)
    g, new_h = model(x, None, score, h)
    assert g.shape == torch.Size([B, C, *D])
    assert new_h.shape == h.shape



from typing import Union

import torch
from torch.nn import functional as F
from torch.nn import Module
from .definitions import DEVICE
import os, re, json
from glob import glob
import numpy as np
import contextlib

class NullEMA:
    """
    An EMA emulator that does nothing so that ema_decay=0 can be supported. 
    """
    def update(self, parameters=None):
        pass

    def copy_to(self, parameters=None):
        pass

    def store(self, parameters=None):
        pass

    def restore(self, parameters=None):
        pass

    def average_parameters(self, parameters=None):
        return contextlib.nullcontext()

    def to(self, device=None, dtype=None):
        pass

    def state_dict(self):
        return {}

    def load_state_dict(self, state_dict):
        pass



def get_activation(activation:str):
    if activation.lower() == "elu":
        return F.elu
    elif activation.lower() == "relu":
        return F.relu
    elif activation.lower() == "tanh":
        return F.tanh
    elif activation.lower() == "swish" or activation.lower() == "silu":
        return F.silu
    elif activation.lower() == "sigmoid":
        return F.sigmoid
    else:
        raise ValueError(f"activation {activation} is not supported")


def load_architecture(
        checkpoints_directory, 
        model: Union[str, Module] = None, 
        dimensions=1, 
        hyperparameters=None, 
        device=DEVICE
        ) -> list[Module, dict]:
    if hyperparameters is None:
        hyperparameters = {}
    if model is None:
        with open(os.path.join(checkpoints_directory, "model_hparams.json"), "r") as f:
            hparams = json.load(f)
        hparams.update(hyperparameters)
        model = hparams.get("architecture", "hourglass")
        if "dimensions" not in hparams.keys():
            hparams["dimensions"] = dimensions
    if isinstance(model, str):
        if model.lower() == "hourglass":
            from rim import Hourglass
            model = Hourglass(**hparams).to(device)
        else:
            raise ValueError(f"{model} not supported")
    else:
        hparams = model.hyperparameters
    paths = glob(os.path.join(checkpoints_directory, "checkpoint*.pt"))
    checkpoints = [int(re.findall('[0-9]+', os.path.split(path)[-1])[-1]) for path in paths]
    if paths:
        try:
            model.load_state_dict(torch.load(paths[np.argmax(checkpoints)], map_location=device))
        except (KeyError, RuntimeError) as e:
            raise e # need to fix what's below
            # # Maybe the RIM instance was used when saving the weights, in which case we hack the loading process
            # from rim.rim import RIM
            # model = RIM(model, **hyperparameters)
            # model.load_state_dict(torch.load(paths[np.argmax(checkpoints)], map_location=device))
            # model = model.model # Remove the RIM wrapping to extract the nn
    return model, hparams 



from setuptools import setup, find_packages

# Read the contents of the README file
with open("long_description.rst", "r") as fh:
    long_description = fh.read()

setup(
    name='torch_rim',
    version='0.2.1',
    author='Alexandre Adam',
    author_email='alexande.adam@umontreal.ca',
    description='A torch implementation of the Recurrent Inference Machine',
    url="https://github.com/AlexandreAdam/torch_rim",
    packages=find_packages(),
    install_requires=[
        'torch>=2.0',
        'score_models>=0.4.4',
        "torch_ema",
        'numpy',
        'tqdm',
        'scipy',
    ],
	python_requires=">=3.8",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.8",
    ],
)




from rim import RIM, Hourglass
import torch

def setup(dim):
    H = 32
    B = 10
    C = 15 # channels
    T = 10
    ch_mult = (2, 2)
    dim = 1
    D = [16]*dim
    x = torch.randn(B, C, *D)
    y = torch.randn_like(x)
    model = Hourglass(C, ch_mult=ch_mult, dimensions=dim)
    score_fn = lambda x, y: torch.randn_like(x)
    energy_fn = lambda x, y: torch.ones([B])
    return x, y, model, score_fn, energy_fn, H, B, C, T, D

def test_rim_hourglass_basic_1d():
    x, y, model, score_fn, energy_fn, H, B, C, T, D = setup(1)
    rim = RIM(D, model, score_fn=score_fn, T=T)
    x_series =  rim(y)
    assert len(x_series) == T
    for x in x_series:
        assert x.shape == torch.Size([B, C, *D])

def test_rim_hourglass_basic_2d():
    x, y, model, score_fn, energy_fn, H, B, C, T, D = setup(2)
    rim = RIM(D, model, score_fn=score_fn, T=T)
    x_series =  rim(y)
    assert len(x_series) == T
    for x in x_series:
        assert x.shape == torch.Size([B, C, *D])

def test_rim_hourglass_energy_1d():
    x, y, model, score_fn, energy_fn, H, B, C, T, D = setup(1)
    rim = RIM(D, model, energy_fn=energy_fn, T=T)
    x_series =  rim(y)
    assert len(x_series) == T
    for x in x_series:
        assert x.shape == torch.Size([B, C, *D])

def test_rim_hourglass_model_init_1d():
    x, y, model, score_fn, energy_fn, H, B, C, T, D = setup(1)
    rim = RIM(D, model, score_fn=score_fn, T=T, initialization_method="model")
    x_series =  rim(y)
    assert len(x_series) == T + 1
    for x in x_series:
        assert x.shape == torch.Size([B, C, *D])


def test_rim_hourglass_link_fn_1d():
    x, y, model, score_fn, energy_fn, H, B, C, T, D = setup(1)
    link_fn = lambda x: 2*x
    inv_link_fn = lambda x: x/2
    rim = RIM(D, model, score_fn=score_fn, T=T, link_function=link_fn, inverse_link_function=inv_link_fn, initialization_method="model")
    x_series =  rim(y)
    assert len(x_series) == T + 1
    for x in x_series:
        assert x.shape == torch.Size([B, C, *D])


def test_rim_hourglass_adam_preprocessing_1d():
    x, y, model, score_fn, energy_fn, H, B, C, T, D = setup(1)
    rim = RIM(D, model, score_fn=score_fn, T=T, score_preprocessing_method="adam")
    x_series =  rim(y)
    assert len(x_series) == T 
    for x in x_series:
        assert x.shape == torch.Size([B, C, *D])

def test_rim_hourglass_arcsinh_1d():
    x, y, model, score_fn, energy_fn, H, B, C, T, D = setup(1)
    rim = RIM(D, model, score_fn=score_fn, T=T, score_preprocessing_method="arcsinh")
    x_series =  rim(y)
    assert len(x_series) == T 
    for x in x_series:
        assert x.shape == torch.Size([B, C, *D])


def test_rim_hourglass_rmsprop_preprocessing_1d():
    x, y, model, score_fn, energy_fn, H, B, C, T, D = setup(1)
    rim = RIM(D, model, score_fn=score_fn, T=T, score_preprocessing_method="rmsprop")
    x_series =  rim(y)
    assert len(x_series) == T
    for x in x_series:
        assert x.shape == torch.Size([B, C, *D])




import torch
from rim import RIM, Hourglass
import os, shutil

def setup(dim):
    H = 32
    B = 10
    C = 15 # channels
    T = 10
    ch_mult = (2, 2)
    dim = 1
    D = [16]*dim
    x = torch.randn(B, C, *D)
    y = torch.randn_like(x)
    model = Hourglass(C, ch_mult=ch_mult, dimensions=dim)
    score_fn = lambda x, y: torch.randn_like(x)
    energy_fn = lambda x, y: torch.ones([B])
    return x, y, model, score_fn, energy_fn, H, B, C, T, D


def test_training():
    x, y, model, score_fn, energy_fn, H, B, C, T, D = setup(1)
    rim = RIM(D, model, score_fn=score_fn, T=T)
    class Dataset(torch.utils.data.Dataset):
        def __len__(self):
            return 2*B
        def __getitem__(self, index):
            return torch.randn(C, *D), torch.randn(C, *D)
    
    dataset = Dataset()
    # Set the hyperparameters and other options for training
    checkpoints_directory = os.path.dirname(os.path.abspath(__file__)) + "/checkpoints"
    learning_rate = 1e-3
    batch_size = B
    epochs = 10
    warmup = 0 # learning rate warmup
    clip = 0. # gradient clipping
    seed = 42

    # Fit the model to the dataset
    losses = rim.fit(
            dataset, 
            learning_rate=learning_rate, 
            checkpoints_directory=checkpoints_directory,
            batch_size=batch_size, 
            checkpoints=1,
            models_to_keep=1,
            epochs=epochs, 
            warmup=warmup, 
            clip=clip, 
            seed=seed
            )
    print(losses)
    # Leave the checkpoints to be removed by last test

    

def test_training_from_checkpoint():
    x, y, model, score_fn, energy_fn, H, B, C, T, D = setup(1)
    checkpoints_directory = os.path.dirname(os.path.abspath(__file__)) + "/checkpoints"
    # Dont pass model, test that we can load it correctly from checkpoints
    rim = RIM(D, checkpoints_directory=checkpoints_directory, score_fn=score_fn, T=T)
    class Dataset(torch.utils.data.Dataset):
        def __len__(self):
            return 2*B
        def __getitem__(self, index):
            return torch.randn(C, *D), torch.randn(C, *D)
    
    dataset = Dataset()
    # Set the hyperparameters and other options for training
    learning_rate = 1e-3
    batch_size = B
    epochs = 10
    warmup = 0 # learning rate warmup
    clip = 0. # gradient clipping
    seed = 42

    # Fit the model to the dataset
    losses = rim.fit(
            dataset, 
            learning_rate=learning_rate, 
            checkpoints_directory=checkpoints_directory,
            models_to_keep=1,
            batch_size=batch_size, 
            epochs=epochs, 
            warmup=warmup, 
            clip=clip, 
            seed=seed
            )
    print(losses)

def test_training_with_ema():
    x, y, model, score_fn, energy_fn, H, B, C, T, D = setup(1)
    checkpoints_directory = os.path.dirname(os.path.abspath(__file__)) + "/checkpoints"
    rim = RIM(D, checkpoints_directory=checkpoints_directory, score_fn=score_fn, T=T)

    class Dataset(torch.utils.data.Dataset):
        def __len__(self):
            return 2*B
        def __getitem__(self, index):
            return torch.randn(C, *D), torch.randn(C, *D)
    
    dataset = Dataset()

    # Set the hyperparameters and other options for training
    learning_rate = 1e-3
    ema_decay=0.99
    batch_size = B
    epochs = 10
    warmup = 0 # learning rate warmup
    clip = 0. # gradient clipping
    seed = 42

    # Fit the model to the dataset
    losses = rim.fit(
            dataset, 
            learning_rate=learning_rate, 
            checkpoints_directory=checkpoints_directory,
            models_to_keep=1,
            ema_decay=ema_decay,
            batch_size=batch_size, 
            epochs=epochs, 
            warmup=warmup, 
            clip=clip, 
            seed=seed
            )

    # When test are finished, remove the checkpoint directory
    shutil.rmtree(checkpoints_directory)
